==> cargo-registry-0.2.1.crate: cargo-registry-0.2.1/src/index.rs <==
//! Interface for working with the crates index.
//!
//! The crates index is stored in a Git repository. Its structure is
//! described [here][index format].
//!
//! This module in particular uses some code from the crates.io
//! codebase -- in particular, from [this file][crates-io]. As it
//! turns out, most operations on the index are fairly
//! straightforward, and there is little to improve over that
//! implementation.
//!
//! [index format]: https://doc.rust-lang.org/cargo/reference/registries.html#index-format
//! [crates-io]: https://github.com/rust-lang/crates.io/blob/master/src/git.rs

use serde::{Deserialize, Serialize};
use std::{
    collections::HashMap,
    fs,
    io::{BufRead, BufReader, Write},
    path::{Path, PathBuf},
};
use tempfile::TempDir;
use tracing::{info, trace};

pub type Result<T> = std::result::Result<T, Error>;

#[derive(Debug, thiserror::Error)]
pub enum Error {
    #[error("IO: {0}")]
    IO(#[from] std::io::Error),

    #[error("Git error: {0}")]
    Git(#[from] git2::Error),

    #[error("Json error: {0}")]
    Json(#[from] serde_json::Error),
}

pub struct Index {
    checkout_path: TempDir,
    repository: git2::Repository,
    private_key: Option<String>,
    basic_auth: Option<(String, String)>,
    name: Option<String>,
    email: Option<String>,
}

impl Index {
    pub fn new(
        url: &str,
        private_key: Option<String>,
        basic_auth: Option<(String, String)>,
        name: Option<String>,
        email: Option<String>,
    ) -> Result<Index> {
        let checkout_path = tempfile::Builder::new().prefix("crates").tempdir()?;
        trace!(dir = %checkout_path.path().display(), "cloning index repository");

        let mut callbacks = git2::RemoteCallbacks::new();
        if let Some(ref private_key) = private_key {
            callbacks.credentials(move |_url, username_from_url, _allowed_types| {
                git2::Cred::ssh_key_from_memory(username_from_url.unwrap(), None, private_key, None)
            });
        }
        if let Some((ref username, ref password)) = basic_auth {
            callbacks.credentials(move |_url, _username_from_url, _allowed_types| {
                git2::Cred::userpass_plaintext(username, password)
            });
        }
        let mut options = git2::FetchOptions::new();
        options.remote_callbacks(callbacks);
        let repository = git2::build::RepoBuilder::new()
            .fetch_options(options)
            .clone(url, checkout_path.path())?;
        trace!("cloned index repository");

        Ok(Index {
            checkout_path,
            repository,
            private_key,
            basic_auth,
            name,
            email,
        })
    }

    fn index_file(&self, name: &str) -> PathBuf {
        self.checkout_path.path().join(relative_index_file(name))
    }

    pub fn add_crate(&self, crate_info: Crate) -> Result<()> {
        info!(%crate_info.name, "add_crate");

        let path = self.index_file(&crate_info.name);
        info!(?path);
        fs::create_dir_all(path.parent().unwrap())?;
        let mut file = fs::OpenOptions::new()
            .append(true)
            .create(true)
            .open(path)?;
        serde_json::to_writer(&mut file, &crate_info)?;
        file.write_all(b"\n")?;

        let message = format!("Add {} version {}", crate_info.name, crate_info.vers);

        self.commit_and_push(&message, &relative_index_file(&crate_info.name))?;

        Ok(())
    }

    pub fn yank(&self, crate_name: &str, version: &str, yanked: bool) -> Result<()> {
        info!(%crate_name, %version, %yanked, "yanking");

        let path = self.index_file(&crate_name);
        let newfile = BufReader::new(fs::File::open(&path)?)
            .lines()
            .map(|line| {
                let line = line?;
                let mut crate_info: Crate = serde_json::from_str(&line)?;
                Ok(if crate_info.vers == version {
                    crate_info.yanked = yanked;
                    serde_json::to_string(&crate_info)?
                } else {
                    line
                })
            })
            .collect::<Result<Vec<_>>>()?
            .join("\n");
        fs::write(path, newfile)?;

        let message = format!(
            "{} {} {}",
            if yanked { "Yanked" } else { "Unyanked" },
            crate_name,
            version
        );
        self.commit_and_push(&message, &relative_index_file(&crate_name))?;

        Ok(())
    }

    fn commit_and_push(&self, message: &str, path: &Path) -> Result<()> {
        info!(%message, "commit and push");

        // git add $file
        let mut index = self.repository.index()?;
        index.add_path(path)?;
        index.write()?;
        let tree_id = index.write_tree()?;
        let tree = self.repository.find_tree(tree_id)?;

        // git commit -m $message
        let head = self.repository.head()?;
        let parent = self.repository.find_commit(head.target().unwrap())?;
        let sig = self.repository.signature().or_else(|_| {
            git2::Signature::now(
                self.name.as_deref().unwrap_or("crates"),
                self.email.as_deref().unwrap_or("localhost"),
            )
        })?;
        self.repository
            .commit(Some("HEAD"), &sig, &sig, message, &tree, &[&parent])?;

        // git push
        let mut origin = self.repository.find_remote("origin")?;
        let mut callbacks = git2::RemoteCallbacks::new();
        if let Some(ref private_key) = self.private_key {
            callbacks.credentials(move |_url, username_from_url, _allowed_types| {
                git2::Cred::ssh_key_from_memory(username_from_url.unwrap(), None, private_key, None)
            });
        }
        if let Some((ref username, ref password)) = self.basic_auth {
            callbacks.credentials(move |_url, _username_from_url, _allowed_types| {
                git2::Cred::userpass_plaintext(username, password)
            });
        }
        callbacks.push_update_reference(|refname, status| match (refname, status) {
            ("refs/heads/master", None) => Ok(()),
            (_, Some(s)) => Err(git2::Error::from_str(&format!(
                "failed to push a ref: {}",
                s
            ))),
            (_, _) => Err(git2::Error::from_str(&format!(
                "pushed the wrong ref: {}",
                refname
            ))),
        });
        let mut opts = git2::PushOptions::new();
        opts.remote_callbacks(callbacks);
        origin.push(&["refs/heads/master"], Some(&mut opts))?;

        Ok(())
    }
}

fn relative_index_file(name: &str) -> PathBuf {
    match name.len() {
        1 => Path::new("1").join(name),
        2 => Path::new("2").join(name),
        3 => Path::new("3").join(&name[..1]).join(name),
        _ => Path::new(&name[..2]).join(&name[2..4]).join(name),
    }
}

#[derive(Deserialize, Serialize)]
pub struct Crate {
    /// The name of the package.
    ///
    /// This must only contain alphanumeric, `-`, or `_` characters.
    pub name: String,

    /// The version of the package this row is describing.
    ///
    /// This must be a valid version number according to the Semantic
    /// Versioning 2.0.0 spec at https://semver.org/.
    pub vers: String,

    /// Array of direct dependencies of the package.
    pub deps: Vec<Dependency>,

    /// A SHA256 checksum of the `.crate` file.
    pub cksum: String,

    /// Set of features defined for the package.
    ///
    /// Each feature maps to an array of features or dependencies it
    /// enables.
    pub features: HashMap<String, Vec<String>>,

    /// Boolean of whether or not this version has been yanked.
    pub yanked: bool,

    /// The `links` string value from the package's manifest, or null
    /// if not specified. This field is optional and defaults to null.
    pub links: Option<String>,
}

#[derive(Deserialize, Serialize)]
pub struct Dependency {
    /// Name of the dependency.
    ///
    /// If the dependency is renamed from the original package name,
    /// this is the new name. The original package name is stored in
    /// the `package` field.
    pub name: String,

    /// The semver requirement for this dependency.
    ///
    /// This must be a valid version requirement defined at
    /// https://github.com/steveklabnik/semver#requirements.
    pub req: String,

    /// Array of features (as strings) enabled for this dependency.
    pub features: Vec<String>,

    /// Boolean of whether or not this is an optional dependency.
    pub optional: bool,

    /// Boolean of whether or not default features are enabled.
    pub default_features: bool,

    /// The target platform for the dependency.
    ///
    /// null if not a target dependency.  Otherwise, a string such as
    /// "cfg(windows)".
    pub target: Option<String>,

    /// The dependency kind.
    ///
    /// "dev", "build", or "normal".
    ///
    /// Note: this is a required field, but a small number of entries
    /// exist in the crates.io index with either a missing or null
    /// `kind` field due to implementation bugs.
    pub kind: DependencyKind,

    /// The URL of the index of the registry where this dependency is
    /// from as a string. If not specified or null, it is assumed the
    /// dependency is in the current registry.
    pub registry: Option<String>,

    /// If the dependency is renamed, this is a string of the actual
    /// package name. If not specified or null, this dependency is not
    /// renamed.
    pub package: Option<String>,
}

#[derive(Deserialize, Serialize)]
#[serde(rename_all = "snake_case")]
pub enum DependencyKind {
    Dev,
    Build,
    Normal,
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn relative_index_files() {
        assert_eq!(relative_index_file("a"), Path::new("1/a"));
        assert_eq!(relative_index_file("ab"), Path::new("2/ab"));
        assert_eq!(relative_index_file("abc"), Path::new("3/a/abc"));
        assert_eq!(relative_index_file("abcd"), Path::new("ab/cd/abcd"));
        assert_eq!(relative_index_file("abcde"), Path::new("ab/cd/abcde"));
    }
}

==> cargo-registry-0.2.2.crate: cargo-registry-0.2.2/src/index.rs <==
//! Interface for working with the crates index.
//!
//! The crates index is stored in a Git repository. Its structure is
//! described [here][index format].
//!
//! This module in particular uses some code from the crates.io
//! codebase -- in particular, from [this file][crates-io]. As it
//! turns out, most operations on the index are fairly
//! straightforward, and there is little to improve over that
//! implementation.
//!
//! [index format]: https://doc.rust-lang.org/cargo/reference/registries.html#index-format
//! [crates-io]: https://github.com/rust-lang/crates.io/blob/master/src/git.rs

use serde::{Deserialize, Serialize};
use std::{
    collections::HashMap,
    fs,
    io::{BufRead, BufReader, Write},
    path::{Path, PathBuf},
};
use tempfile::TempDir;
use tracing::{info, trace};

pub type Result<T> = std::result::Result<T, Error>;

#[derive(Debug, thiserror::Error)]
pub enum Error {
    #[error("IO: {0}")]
    IO(#[from] std::io::Error),

    #[error("Git error: {0}")]
    Git(#[from] git2::Error),

    #[error("Json error: {0}")]
    Json(#[from] serde_json::Error),

    #[error("Failed to clone repo after too many retries")]
    Retry(#[from] retry::Error<git2::Error>),
}

pub struct Index {
    checkout_path: TempDir,
    repository: git2::Repository,
    private_key: Option<String>,
    basic_auth: Option<(String, String)>,
    name: Option<String>,
    email: Option<String>,
}

impl Index {
    pub fn new(
        url: &str,
        private_key: Option<String>,
        basic_auth: Option<(String, String)>,
        name: Option<String>,
        email: Option<String>,
    ) -> Result<Index> {
        let checkout_path = tempfile::Builder::new().prefix("crates").tempdir()?;
        trace!(dir = %checkout_path.path().display(), "cloning index repository");

        let repository = retry::retry(retry::delay::Fibonacci::from_millis(1000).take(10), || {
            let mut callbacks = git2::RemoteCallbacks::new();
            if let Some(ref private_key) = private_key {
                callbacks.credentials(move |_url, username_from_url, _allowed_types| {
                    git2::Cred::ssh_key_from_memory(
                        username_from_url.unwrap(),
                        None,
                        private_key,
                        None,
                    )
                });
            }
            if let Some((ref username, ref password)) = basic_auth {
                callbacks.credentials(move |_url, _username_from_url, _allowed_types| {
                    git2::Cred::userpass_plaintext(username, password)
                });
            }

            let mut options = git2::FetchOptions::new();
            options.remote_callbacks(callbacks);
            let repo = git2::build::RepoBuilder::new()
                .fetch_options(options)
                .clone(url, checkout_path.path())?;
            Ok(repo)
        })?;
        trace!("cloned index repository");

        Ok(Index {
            checkout_path,
            repository,
            private_key,
            basic_auth,
            name,
            email,
        })
    }

    fn index_file(&self, name: &str) -> PathBuf {
        self.checkout_path.path().join(relative_index_file(name))
    }

    pub fn add_crate(&self, crate_info: Crate) -> Result<()> {
        info!(%crate_info.name, "add_crate");

        let path = self.index_file(&crate_info.name);
        info!(?path);
        fs::create_dir_all(path.parent().unwrap())?;
        let mut file = fs::OpenOptions::new()
            .append(true)
            .create(true)
            .open(path)?;
        serde_json::to_writer(&mut file, &crate_info)?;
        file.write_all(b"\n")?;

        let message = format!("Add {} version {}", crate_info.name, crate_info.vers);

        self.commit_and_push(&message, &relative_index_file(&crate_info.name))?;

        Ok(())
    }

    pub fn yank(&self, crate_name: &str, version: &str, yanked: bool) -> Result<()> {
        info!(%crate_name, %version, %yanked, "yanking");

        let path = self.index_file(&crate_name);
        let newfile = BufReader::new(fs::File::open(&path)?)
            .lines()
            .map(|line| {
                let line = line?;
                let mut crate_info: Crate = serde_json::from_str(&line)?;
                Ok(if crate_info.vers == version {
                    crate_info.yanked = yanked;
                    serde_json::to_string(&crate_info)?
                } else {
                    line
                })
            })
            .collect::<Result<Vec<_>>>()?
            .join("\n");
        fs::write(path, newfile)?;

        let message = format!(
            "{} {} {}",
            if yanked { "Yanked" } else { "Unyanked" },
            crate_name,
            version
        );
        self.commit_and_push(&message, &relative_index_file(&crate_name))?;

        Ok(())
    }

    fn commit_and_push(&self, message: &str, path: &Path) -> Result<()> {
        info!(%message, "commit and push");

        // git add $file
        let mut index = self.repository.index()?;
        index.add_path(path)?;
        index.write()?;
        let tree_id = index.write_tree()?;
        let tree = self.repository.find_tree(tree_id)?;

        // git commit -m $message
        let head = self.repository.head()?;
        let parent = self.repository.find_commit(head.target().unwrap())?;
        let sig = self.repository.signature().or_else(|_| {
            git2::Signature::now(
                self.name.as_deref().unwrap_or("crates"),
                self.email.as_deref().unwrap_or("localhost"),
            )
        })?;
        self.repository
            .commit(Some("HEAD"), &sig, &sig, message, &tree, &[&parent])?;

        // git push
        let mut origin = self.repository.find_remote("origin")?;
        let mut callbacks = git2::RemoteCallbacks::new();
        if let Some(ref private_key) = self.private_key {
            callbacks.credentials(move |_url, username_from_url, _allowed_types| {
                git2::Cred::ssh_key_from_memory(username_from_url.unwrap(), None, private_key, None)
            });
        }
        if let Some((ref username, ref password)) = self.basic_auth {
            callbacks.credentials(move |_url, _username_from_url, _allowed_types| {
                git2::Cred::userpass_plaintext(username, password)
            });
        }
        callbacks.push_update_reference(|refname, status| match (refname, status) {
            ("refs/heads/master", None) => Ok(()),
            (_, Some(s)) => Err(git2::Error::from_str(&format!(
                "failed to push a ref: {}",
                s
            ))),
            (_, _) => Err(git2::Error::from_str(&format!(
                "pushed the wrong ref: {}",
                refname
            ))),
        });
        let mut opts = git2::PushOptions::new();
        opts.remote_callbacks(callbacks);
        origin.push(&["refs/heads/master"], Some(&mut opts))?;

        Ok(())
    }
}

fn relative_index_file(name: &str) -> PathBuf {
    match name.len() {
        1 => Path::new("1").join(name),
        2 => Path::new("2").join(name),
        3 => Path::new("3").join(&name[..1]).join(name),
        _ => Path::new(&name[..2]).join(&name[2..4]).join(name),
    }
}

#[derive(Deserialize, Serialize)]
pub struct Crate {
    /// The name of the package.
    ///
    /// This must only contain alphanumeric, `-`, or `_` characters.
    pub name: String,

    /// The version of the package this row is describing.
    ///
    /// This must be a valid version number according to the Semantic
    /// Versioning 2.0.0 spec at https://semver.org/.
    pub vers: String,

    /// Array of direct dependencies of the package.
    pub deps: Vec<Dependency>,

    /// A SHA256 checksum of the `.crate` file.
    pub cksum: String,

    /// Set of features defined for the package.
    ///
    /// Each feature maps to an array of features or dependencies it
    /// enables.
    pub features: HashMap<String, Vec<String>>,

    /// Boolean of whether or not this version has been yanked.
    pub yanked: bool,

    /// The `links` string value from the package's manifest, or null
    /// if not specified. This field is optional and defaults to null.
    pub links: Option<String>,
}

#[derive(Deserialize, Serialize)]
pub struct Dependency {
    /// Name of the dependency.
    ///
    /// If the dependency is renamed from the original package name,
    /// this is the new name. The original package name is stored in
    /// the `package` field.
    pub name: String,

    /// The semver requirement for this dependency.
    ///
    /// This must be a valid version requirement defined at
    /// https://github.com/steveklabnik/semver#requirements.
    pub req: String,

    /// Array of features (as strings) enabled for this dependency.
    pub features: Vec<String>,

    /// Boolean of whether or not this is an optional dependency.
    pub optional: bool,

    /// Boolean of whether or not default features are enabled.
    pub default_features: bool,

    /// The target platform for the dependency.
    ///
    /// null if not a target dependency.  Otherwise, a string such as
    /// "cfg(windows)".
    pub target: Option<String>,

    /// The dependency kind.
    ///
    /// "dev", "build", or "normal".
    ///
    /// Note: this is a required field, but a small number of entries
    /// exist in the crates.io index with either a missing or null
    /// `kind` field due to implementation bugs.
    pub kind: DependencyKind,

    /// The URL of the index of the registry where this dependency is
    /// from as a string. If not specified or null, it is assumed the
    /// dependency is in the current registry.
    pub registry: Option<String>,

    /// If the dependency is renamed, this is a string of the actual
    /// package name. If not specified or null, this dependency is not
    /// renamed.
    pub package: Option<String>,
}

#[derive(Deserialize, Serialize)]
#[serde(rename_all = "snake_case")]
pub enum DependencyKind {
    Dev,
    Build,
    Normal,
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn relative_index_files() {
        assert_eq!(relative_index_file("a"), Path::new("1/a"));
        assert_eq!(relative_index_file("ab"), Path::new("2/ab"));
        assert_eq!(relative_index_file("abc"), Path::new("3/a/abc"));
        assert_eq!(relative_index_file("abcd"), Path::new("ab/cd/abcd"));
        assert_eq!(relative_index_file("abcde"), Path::new("ab/cd/abcde"));
    }
}
